/*
 * op.c - Operators
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */


#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "util.h"
#include "error.h"
#include "data.h"
#include "op.h"


#define UNTYPED_NUMBER(d) ((d).type == dt_unum || (d).type == dt_fnum)
#define TYPED_NUMBER(d) \
  ((d).type == dt_rate || (d).type == dt_prate || \
   (d).type == dt_size || (d).type == dt_psize || (d).type == dt_time)
#define TYPES2(a,b,t1,t2) \
  (((a).type == (t1) && (b).type == (t2)) || \
   ((a).type == (t2) && (b).type == (t1)))

#define __OP(type,a,op,b) \
  (data_##type(data_convert(a,dt_##type).u.type op \
  data_convert(b,dt_##type).u.type))
#define UNUM_OP(a,op,b) __OP(unum,a,op,b)
#define FNUM_OP(a,op,b) __OP(fnum,a,op,b)
#define ANYNUM_OP(a,op,b) (all_unum(a,b) ? UNUM_OP(a,op,b) : FNUM_OP(a,op,b))



static DATA make_op(OP *op,DATA_TYPE type)
{
    DATA d;

    d.type = type;
    d.op = op;
    return d;
}


static void __attribute__((noreturn)) type_error(OP *op)
{
    yyerrorf("invalid types for operator \"%s\"",op->dsc->name);
}


/* ----- +,- --------------------------------------------------------------- */


static DATA additive_eval(OP *op)
{
    int plus = *op->dsc->name == '+';
    DATA d;

    if (op->b.type == dt_unum && !op->b.op && !op->b.u.unum) return op->a;
    if (plus && op->a.type == dt_unum && !op->a.op && !op->a.u.unum)
	return op->b;
    if (UNTYPED_NUMBER(op->a) && UNTYPED_NUMBER(op->b)) {
	if (op->a.op || op->b.op)
	    return make_op(op,all_unum(op->a,op->b) ? dt_unum : dt_fnum);
	return plus ? ANYNUM_OP(op->a,+,op->b) : ANYNUM_OP(op->a,-,op->b);
    }
    if (op->a.type == op->b.type && TYPED_NUMBER(op->a)) {
	if (op->a.op || op->b.op) return make_op(op,op->a.type);
	d = op->a;
	if (plus) d.u.fnum += op->b.u.fnum;
	else d.u.fnum -= op->b.u.fnum;
	return d;
    }
    type_error(op);
}


OP_DSC op_plus = {
    .name = "+",
    .eval = additive_eval,
};

OP_DSC op_minus = {
    .name = "-",
    .eval = additive_eval,
};


/* ----- * ----------------------------------------------------------------- */


static DATA mult_eval(OP *op)
{
    DATA d;

    if (UNTYPED_NUMBER(op->a) && UNTYPED_NUMBER(op->b)) {
	if (op->a.op || op->b.op)
	    return make_op(op,all_unum(op->a,op->b) ? dt_unum : dt_fnum);
	return ANYNUM_OP(op->a,*,op->b);
    }
    if ((UNTYPED_NUMBER(op->a) && TYPED_NUMBER(op->b)) ||
      (UNTYPED_NUMBER(op->b) && TYPED_NUMBER(op->a))) {
	if (op->a.op || op->b.op)
	    return make_op(op,TYPED_NUMBER(op->a) ? op->a.type : op->b.type);
	if (TYPED_NUMBER(op->a)) {
	    d = op->a;
	    d.u.fnum *= data_convert(op->b,dt_fnum).u.fnum;
	}
	else {
	    d = op->b;
	    d.u.fnum *= data_convert(op->a,dt_fnum).u.fnum;
	}
	return d;
    }
    d.type = dt_none;
    d.u.fnum = op->a.u.fnum;
    if (TYPES2(op->a,op->b,dt_rate,dt_time)) {
	d.type = dt_size;
	d.u.fnum /= 8.0;
    }
    if (TYPES2(op->a,op->b,dt_prate,dt_time)) d.type = dt_psize;
#if 0 /* maybe later */
    if (TYPES2(op->a,op->b,dt_inv_rate,dt_rate)) d.type = dt_time;
    if (TYPES2(op->a,op->b,dt_inv_prate,dt_prate)) d.type = dt_time;
#endif
    if (d.type == dt_none) type_error(op);
    if (op->a.op || op->b.op) return make_op(op,d.type);
    d.op = NULL;
    d.u.fnum *= op->b.u.fnum;
    return d;
}


OP_DSC op_mult = {
    .name = "*",
    .eval = mult_eval,
};


/* ----- / ----------------------------------------------------------------- */


static DATA div_eval(OP *op)
{
    DATA d;

    if (UNTYPED_NUMBER(op->a) && UNTYPED_NUMBER(op->b)) {
	if (op->a.op || op->b.op)
	    return make_op(op,all_unum(op->a,op->b) ? dt_unum : dt_fnum);
	return ANYNUM_OP(op->a,/,op->b);
    }
    if (TYPED_NUMBER(op->a) && UNTYPED_NUMBER(op->b)) {
	if (op->a.op || op->b.op) return make_op(op,op->a.type);
	d = op->a;
	d.u.fnum /= data_convert(op->b,dt_fnum).u.fnum;
	return d;
    }
    d.type = dt_none;
    d.u.fnum = op->a.u.fnum;
    if (op->a.type == dt_size && op->b.type == dt_time) {
	d.type = dt_rate;
	d.u.fnum *= 8.0;
    }
    if (op->a.type == dt_psize && op->b.type == dt_time) d.type = dt_prate;
#if 0 /* maybe later */
    if (TYPES2(op->a,op->b,dt_inv_rate,dt_rate)) d.type = dt_time;
    if (TYPES2(op->a,op->b,dt_inv_prate,dt_prate)) d.type = dt_time;
#endif
    if (d.type == dt_none) type_error(op);
    if (op->a.op || op->b.op) return make_op(op,d.type);
    d.op = NULL;
    d.u.fnum /= op->b.u.fnum;
    return d;
}


OP_DSC op_div = {
    .name = "/",
    .eval = div_eval,
};


/* ----- ~,!,- ------------------------------------------------------------- */


static DATA unum1_eval(OP *op)
{
    DATA d;

    if ((op->a.type != dt_unum && *op->dsc->name != '-') ||
      (!UNTYPED_NUMBER(op->a) && !TYPED_NUMBER(op->a)))
	type_error(op);
    if (op->a.op) return make_op(op,dt_unum);
    /* @@@ convert fnum to unum ? */
    switch (*op->dsc->name) {
	case '~':
	    return data_unum(~op->a.u.unum);
	case '!':
	    return data_unum(!op->a.u.unum);
	case '-':
	    if (op->a.type == dt_unum) return data_unum(-op->a.u.unum);
	    d = op->a;
	    d.u.fnum = -d.u.fnum;
	    return d;
	default:
	    abort();
    }
}


OP_DSC op_not = {
    .name = "~",
    .eval = unum1_eval,
};


OP_DSC op_logical_not = {
    .name = "!",
    .eval = unum1_eval,
};


OP_DSC op_unary_minus = {
    .name = "-",
    .eval = unum1_eval,
};


/* ----- %, &, |, ^, <<, >>, ||, && ---------------------------------------- */


/* Macros are fun ;-) */


static DATA unum2_eval(OP *op);


#define UNUM2_OPS(m) \
  m(mod,%) \
  m(and,&) m(or,|) m(xor,^) \
  m(shift_left,<<) m(shift_right,>>) \
  m(logical_or,||) m(logical_and,&&)

#define _DECLARATION(n,op) \
  static const char op_name_##n[] = #op; \
  OP_DSC op_##n = { \
    .name = op_name_##n, \
    .eval = unum2_eval, \
  };

UNUM2_OPS(_DECLARATION)

#undef _DECLARATION


#define _EVALUATION(n,op) \
  if (name == op_name_##n) return a op b;

static uint32_t unum2_op(const char *name,uint32_t a,uint32_t b)
{
    /*
     * According to K&R 2nd ed., I should be able to use switch here, but
     * gcc 2.91.66 / egcs 1.1.2 won't let me.
     */
    UNUM2_OPS(_EVALUATION)
    abort();
}

#undef _EVALUATION


static DATA unum2_eval(OP *op)
{
    /*
     * use classes for short-cut evaluation in "if"
     */
    if ((op->a.type != dt_unum && op->a.type != dt_decision) ||
      (op->b.type != dt_unum && op->b.type != dt_decision))
	type_error(op);
    if (op->a.op || op->b.op || op->a.type == dt_decision ||
      op->b.type == dt_decision)
	return make_op(op,dt_unum);
    /* @@@ convert fnum to unum ? */
    return data_unum(unum2_op(op->dsc->name,op->a.u.unum,op->b.u.unum));
}


/* ----- ==, !=, >, >=, <, <= ---------------------------------------------- */


static DATA rel_eval(OP *op);


#define REL_OPS(m) \
  m(eq,==) m(ne,!=) m(gt,>) m(ge,>=) m(lt,<) m(le,<=)

#define _DECLARATION(n,op) \
  static const char op_name_##n[] = #op; \
  OP_DSC op_##n = { \
    .name = op_name_##n, \
    .eval = rel_eval, \
  };

REL_OPS(_DECLARATION)

#undef _DECLARATION


#define _EVALUATION(n,op) \
  if (name == op_name_##n) return a op b;

static uint32_t rel_unum_op(const char *name,uint32_t a,uint32_t b)
{
    REL_OPS(_EVALUATION)
    abort();
}


static uint32_t rel_fnum_op(const char *name,double a,double b)
{
    REL_OPS(_EVALUATION)
    abort();
}

#undef _EVALUATION


static DATA rel_eval(OP *op)
{
    if (UNTYPED_NUMBER(op->a) && UNTYPED_NUMBER(op->b)) {
	if (op->a.op || op->b.op) return make_op(op,dt_unum);
	if (all_unum(op->a,op->b))
	    return data_unum(rel_unum_op(op->dsc->name,
	      op->a.u.unum,op->b.u.unum));
	return data_unum(rel_fnum_op(op->dsc->name,
	  data_convert(op->a,dt_fnum).u.fnum,
	  data_convert(op->b,dt_fnum).u.fnum));
    }
    if (op->a.type != op->b.type || !TYPED_NUMBER(op->a)) type_error(op);
    if (op->a.op || op->b.op) return make_op(op,dt_unum);
    return data_unum(rel_fnum_op(op->dsc->name,op->a.u.fnum,op->b.u.fnum));
}


/* ----- : : --------------------------------------------------------------- */


static DATA mask_eval(OP *op)
{
    uint32_t mask;
    int bits;

    if (op->a.type != dt_unum ||
      (op->b.type != dt_unum && op->b.type != dt_none) ||
      (op->c.type != dt_unum && op->c.type != dt_none))
	type_error(op);
    if (op->b.op || op->c.op) return make_op(op,dt_unum);
    bits = (op->b.type == dt_none ? 0 : op->b.u.unum)+
      (op->c.type == dt_none ? 0 : op->c.u.unum);
    if (bits > 32) yyerror("mask exceeds 32 bit range");
    mask = (1 << (op->c.type == dt_unum ? op->c : op->b).u.unum)-1;
    if (op->b.type == dt_unum) mask <<= 32-bits;
    return op_binary(&op_and,op->a,data_unum(mask));
}


OP_DSC op_mask = {
    .name = "mask (::)",
    .eval = mask_eval,
};


/* ----- Packet data access ------------------------------------------------ */


static DATA access_eval(OP *op)
{
    /* type checking done by parser */
    return make_op(op,dt_unum);
}


OP_DSC op_access = {
    .name = "access",
    .eval = access_eval,
};


static DATA offset_eval(OP *op)
{
    /* only used internally - no type checking needed */
    return make_op(op,op->a.type);
}


OP_DSC op_offset = {
    .name = "offset",
    .eval = offset_eval,
};


/* ----- Policing ---------------------------------------------------------- */


static DATA conform_eval(OP *op)
{
    if (op->a.type != dt_police) type_error(op);
    return make_op(op,dt_unum);
}


OP_DSC op_conform = {
    .name = "conform",
    .eval = conform_eval,
};


static DATA count_eval(OP *op)
{
    if (op->a.type != dt_police) type_error(op);
    return make_op(op,dt_unum);
}


OP_DSC op_count = {
    .name = "count",
    .eval = count_eval,
};


/* ----- Preconditions ----------------------------------------------------- */


static DATA precond_eval(OP *op)
{
    /* only used internally - no type checking needed */
    return make_op(op,op->a.type);
}


OP_DSC op_precond = {
    .name = "precond",
    .eval = precond_eval,
};


/* ----- Operator construction --------------------------------------------- */


static OP *alloc_op(void)
{
    OP *op = alloc_t(OP);

    memset(op,0,sizeof(*op));
    return op;
}


static DATA eval_op(OP *op)
{
    DATA d;

    d = op->dsc->eval(op);
    if (!d.op) free(op);
    return d;
}


DATA op_unary(const OP_DSC *dsc,DATA a)
{
    OP *op = alloc_op();

    op->dsc = dsc;
    op->a = a;
    return eval_op(op);
}


DATA op_binary(const OP_DSC *dsc,DATA a,DATA b)
{
    OP *op = alloc_op();

    op->dsc = dsc;
    op->a = a;
    op->b = b;
    return eval_op(op);
}


DATA op_ternary(const OP_DSC *dsc,DATA a,DATA b,DATA c)
{
    OP *op = alloc_op();

    op->dsc = dsc;
    op->a = a;
    op->b = b;
    op->c = c;
    return eval_op(op);
}


/* ----- Debugging output -------------------------------------------------- */


struct _prev{
    int indent;
    const struct _prev *prev;
};


static void parents(FILE *file,const struct _prev *curr)
{
    int i;

    if (!curr) return;
    if (curr->prev) {
	parents(file,curr->prev);
	fprintf(file," |");
    }
    for (i = curr->indent-(curr->prev ? curr->prev->indent+2 : 0); i; i--)
	fputc(' ',file);
}


static void do_indent(FILE *file,char branch,int indent,
  const struct _prev *prev)
{
    int i;

    parents(file,prev);
    fprintf(file," %c",branch);
    for (i = indent-(prev ? prev->indent : 0)-2; i; i--) fputc('-',file);
}


static void do_print_expr(FILE *file,DATA d,int indent,
  const struct _prev *prev)
{
    struct _prev new_prev;
    int new_indent;

    new_prev.indent = indent;
    new_prev.prev = prev;
    if (!d.op) {
	fputc(' ',file);
	print_data(file,d);
	fputc('\n',file);
	return;
    }
    fprintf(file,"[%s]--",d.op->dsc->name);
    new_indent = indent+strlen(d.op->dsc->name)+4;
    if (d.op->b.type == dt_none) do_print_expr(file,d.op->a,new_indent,prev);
    else {
	do_print_expr(file,d.op->a,new_indent,&new_prev);
	if (d.op->c.type == dt_none) {
	    do_indent(file,'`',new_indent,&new_prev);
	    do_print_expr(file,d.op->b,new_indent,prev);
	}
	else {
	    do_indent(file,'+',new_indent,&new_prev);
	    do_print_expr(file,d.op->b,new_indent,&new_prev);
	    do_indent(file,'`',new_indent,&new_prev);
	    do_print_expr(file,d.op->c,new_indent,prev);
	}
    }
}


void print_expr(FILE *file,DATA d)
{
    do_print_expr(file,d,0,NULL);
}


void debug_expr(const char *title,DATA d)
{
    int n;

    if (!debug) return;
    n = printf("\n===== %s ",title);
    for (n = 79-n; n; n--) putchar('=');
    printf("\n\n");
    print_expr(stdout,d);
}
